#!python

import shutil
import os 
from pathlib import Path
import traceback
import yaml
import re

from dtiplayground.config import INFO as info
import dtiplayground
import dtiplayground.dmri.common as common
import dtiplayground.dmri.common.module as module
from dtiplayground.dmri.common.appbase import AppBase

import dtiplayground.dmri.preprocessing as preprocessing
logger=common.logger.write 
color= common.Color

class DMRIPrepApp(AppBase):

    ######### Mandatory Implementations Begins
    def __init__(self, config_root, app_name = 'dmriprep', *args, **kwargs): 
        super().__init__(config_root,app_name,*args,**kwargs)

    def getAppInfoImpl(self):
        return self.app

    def initializeImpl(self, options):

        home_dir = Path(self.app['application_dir'])

        user_module_dir=home_dir.parent.joinpath('modules/{}'.format(self.app['application'])).absolute()
        user_module_dir.mkdir(parents=True,exist_ok=True)
        user_tools_param_dir=home_dir.joinpath('parameters').absolute()
        user_tools_param_dir.mkdir(parents=True,exist_ok=True)

        template_filename=home_dir.joinpath("protocol_template.yml").absolute()
        source_template_path=Path(preprocessing.__file__).parent.joinpath("templates/protocol_template.yml")
        protocol_template=yaml.safe_load(open(source_template_path,'r'))
        config_filename=home_dir.joinpath("config.yml")
        environment_filename=home_dir.joinpath("environment.yml")
        ## make configuration file (config.yml)
        config={"user_module_directories": [str(user_module_dir)],"protocol_template_path" : 'protocol_template.yml'}
        yaml.dump(protocol_template,open(template_filename,'w'))
        yaml.dump(config,open(config_filename,'w'))
        logger("Config file written to : {}".format(str(config_filename)),color.INFO)
        ## copy default software path
        software_filename=home_dir.joinpath('software_paths.yml')
        self._update_software_paths(self.globalvars)
        logger("Software path file is written to : {}".format(str(software_filename)),color.INFO)
        system_module_paths = [Path(dtiplayground.__file__).resolve().parent.joinpath('dmri/preprocessing/modules')]
        modules=module.load_modules(system_module_paths = system_module_paths, user_module_paths=config['user_module_directories'])
        environment=module.generate_module_envionrment(modules,str(home_dir))
        yaml.dump(environment,open(environment_filename,'w'))
        logger("Environment file written to : {}".format(str(environment_filename)),color.INFO) 
        return True

    #######################
    #### default command
    #######################

    def runImpl(self,_options):
        return self.runQC(_options)
    ######### Mandatory Implementations Ends###############################################
   
    #######################
    #### commands
    #######################

    def runQC(self,_options):

        @self.after_initialized
        def _run(_options):
            _options.setdefault('output_file_base', None)
            _options.setdefault('baseline_threshold', 10)
            _options.setdefault('execution_id', common.get_uuid())
            _options.setdefault('output_format', None)
            _options.setdefault('global_variables',{})
            _options.setdefault('no_output_image', False)

            options={
                "config_dir" : self.app['application_dir'],
                "input_image_paths" : _options['input_image_paths'],
                "protocol_path" : _options['protocol_path'],
                "output_dir" : _options['output_dir'],
                "default_protocols":_options['default_protocols'],
                "num_threads":_options['num_threads'],
                "execution_id":_options['execution_id'],
                "baseline_threshold" : _options['baseline_threshold'],
                "output_format" : _options['output_format'],
                "output_file_base" : _options['output_file_base'],
                "no_output_image" : _options['no_output_image'],
                "global_variables" : _options['global_variables']
            }

            if options['output_format'] is not None:
                options['output_format']=options['output_format'].lower()
            ## load config file and run pipeline
            config,environment = self._load_configurations()
            template_path=Path(options['config_dir']).joinpath(config['protocol_template_path'])
            template=yaml.safe_load(open(template_path,'r'))
            proto=preprocessing.protocols.Protocols(options['config_dir'], global_vars=options['global_variables'])
            proto.loadImages(options['input_image_paths'],b0_threshold=options['baseline_threshold'])
            if options['output_dir'] is None:
                raise Exception("Output directory is missing")
            else:
                proto.setOutputDirectory(options['output_dir'])
            if options['default_protocols'] is not None:
                if len(options['default_protocols'])==0:
                    options['default_protocols']=None
                proto.makeDefaultProtocols(options['default_protocols'],template=template,options=options)
            elif options['protocol_path'] is not None:
                proto.loadProtocols(options["protocol_path"])
            else :
                proto.makeDefaultProtocols(options['default_protocols'],template=template,options=options)
            if options['num_threads'] is not None:
                proto.setNumThreads(options['num_threads'])
            Path(options['output_dir']).mkdir(parents=True,exist_ok=True)
            logfilename=str(Path(options['output_dir']).joinpath('log.txt').absolute())
            common.logger.setLogfile(logfilename)  
            logger("\r----------------------------------- QC Begins ----------------------------------------\n")
            res=proto.runPipeline(options=options)
            logger("\r----------------------------------- QC Done ----------------------------------------\n")
            return res 

        return _run(_options)

    def makeProtocols(self, _options):

        @self.after_initialized
        def make_protocol(options):
            options.setdefault('baseline_threshold', 10)
            options.setdefault('output_format', None)
            options.setdefault('global_variables',{})
            options.setdefault('no_output_image', False)
            options={
                "config_dir" : self.app['application_dir'],
                "input_image_paths" : options['input_images'],
                "module_list": options['module_list'],
                "output_path" : options['output'],
                "baseline_threshold" : options['b0_threshold'],
                "output_format" : options['output_format'],
                "no_output_image" : options['no_output_image'],
                "global_variables" : options['global_variables']
            }
            ## load config file
            config,environment = self._load_configurations()
            template_path=Path(options['config_dir']).joinpath(config['protocol_template_path'])
            template=yaml.safe_load(open(template_path,'r'))
            proto=preprocessing.protocols.Protocols(options['config_dir'],global_vars=options['global_variables'])
            proto.loadImages(options['input_image_paths'],b0_threshold=options['baseline_threshold'])
            if options['module_list'] is not None and  len(options['module_list'])==0:
                    options['module_list']=None
            proto.makeDefaultProtocols(options['module_list'],template=template,options=options)
            outstr=yaml.dump(proto.getProtocols())
            # print(outstr)
            if options['output_path'] is not None:
                open(options['output_path'],'w').write(outstr)
                logger("Protocol file has been writte to : {}".format(options['output_path']),color.OK)

            return True 
        return make_protocol(_options)

    def remove_module(self,_options):

        @self.after_initialized
        def _remove_module(args):
            home_dir=Path(self.app['application_dir'])
            module_name = args['name']
            config_filename=home_dir.joinpath("config.yml")
            environment_filename=home_dir.joinpath("environment.yml")
            config=yaml.safe_load(open(config_filename,'r'))
            
            module_root_dir = Path(config['user_module_directories'][0])
            module_dir = module_root_dir.joinpath(module_name)
            if module_dir.exists():
                logger("Removing the module : {} at {}".format(module_name, module_dir.__str__()), color.PROCESS)
                shutil.rmtree(module_dir)
                logger("Module {} has been successfully removed".format(module_name), color.OK)
            else:
                logger("There is no such module in {}".format(module_root_dir.__str__()), color.ERROR)

            return True

        return _remove_module(_options)

    def add_module(self,_options):
        @self.after_initialized
        def _add_module(args):
            home_dir=Path(self.app['application_dir'])
            module_name = args['name']
            base_module_name = args['base_module']
            config_filename=home_dir.joinpath("config.yml")
            environment_filename=home_dir.joinpath("environment.yml")
            config=yaml.safe_load(open(config_filename,'r'))
            system_module_root_dir = Path(dtiplayground.__file__).resolve().parent.joinpath("dmri/preprocessing/modules")
            user_module_root_dir = Path(config['user_module_directories'][0])
            module_dir = user_module_root_dir.joinpath(module_name)

            if self._check_if_module_exists(module_name):
                logger("Module {} exists in {} or {}".format(module_name, user_module_root_dir.__str__(), system_module_root_dir.__str__()), color.ERROR)
                return False

            if base_module_name is None: ### NEW module 
                logger("Making directories and files ... ", color.PROCESS)
                module_dir.mkdir(parents=True,exist_ok=False)
                module_template_fn =Path(dtiplayground.__file__).resolve().parent.joinpath("dmri/preprocessing/templates/module_template.py")
                module_template = open(module_template_fn,'r').read()
                module_data_fn = module_template_fn.parent.joinpath('module_template.yml')
                module_readme_fn = module_template_fn.parent.joinpath('module_template.md')
                module_data = open(module_data_fn,'r').read()
                module_readme = open(module_readme_fn,'r').read()
                ### replacing variable
                regex=r"@MODULENAME@"
                module_py = re.sub(regex,module_name,module_template,0)
                module_yml = re.sub(regex,module_name,module_data,0)
                module_md = re.sub(regex,module_name, module_readme, 0)

                ### write down files into user module dir

                init_fn = module_dir.joinpath("__init__.py")
                with open(init_fn,'w') as f:
                    f.write("")
                out_py_fn = module_dir.joinpath("{}.py".format(module_name))
                with open(out_py_fn,'w') as f:
                    f.write(module_py)
                out_yml_fn = module_dir.joinpath("{}.yml".format(module_name))
                with open(out_yml_fn,'w') as f:
                    f.write(module_yml)
                out_md_fn = module_dir.joinpath("README.md")
                with open(out_md_fn,'w') as f:
                    f.write(module_md)
                
                logger("Module is added, please implement the module {} in {}".format(module_name, module_dir.__str__()), color.OK)
            else: ### copy base module to new one (new name)
                base_module_dir= None
                if system_module_root_dir.joinpath(base_module_name).exists():
                    base_module_dir = system_module_root_dir.joinpath(base_module_name)
                    logger("Module {} found in system module directory".format(base_module_name),color.OK)
                elif user_module_root_dir.joinpath(base_module_name).exists():
                    base_module_dir = user_module_root_dir.joinpath(base_module_name)
                    logger("Module {} found in user module directory".format(base_module_name),color.OK)
                else:
                    logger("Module {} NOT found in either system or user module directory".format(base_module_name),color.ERROR)
                    return False

                logger("Generating {} at {}".format(module_name, module_dir.__str__()),color.PROCESS)
                shutil.copytree(base_module_dir, module_dir)
                if module_dir.joinpath('__pycache__').exists():
                    shutil.rmtree(module_dir.joinpath('__pycache__'))
                flist = list(filter(lambda x: '__pycache__' not in x.__str__(), module_dir.glob("**/*")))
                for fn in flist: ### changing module name in contents (may not perfect)
                    out_content=None
                    with open(fn,'r') as f:
                        ### change class name to new module name
                        content = f.read()
                        out_content = content.replace(base_module_name, module_name)
                        out_content = out_content.replace(base_module_name.replace("_"," "), module_name.replace("_"," "))
                    with open(fn,'w') as f:
                        f.write(out_content)

                    nameparts = fn.name.split('.')
                    if nameparts[0] == base_module_name:
                        nameparts[0] = module_name
                    out_fn = ".".join(nameparts)
                    fn.rename(module_dir.joinpath(out_fn))
                logger("Module {} cloned from {} has been generated at {}".format(module_name, base_module_name, module_dir.__str__()),color.OK)
                if args['edit']:
                    command=["vi",module_dir.joinpath("{}.py".format(module_name))]
                    subprocess.run(command)
                return True

        return _add_module(_options)

    #############        
    ### utilities
    ##############

    def _check_if_module_exists(self,module_name):
        home_dir=Path(self.app['application_dir'])

        config_filename=home_dir.joinpath("config.yml")
        environment_filename=home_dir.joinpath("environment.yml")
        config=yaml.safe_load(open(config_filename,'r'))
        
        system_module_root_dir = Path(dtiplayground.__file__).resolve().parent.joinpath("dmri/preprocessing/modules")
        user_module_root_dir = Path(config['user_module_directories'][0])
        if system_module_root_dir.joinpath(module_name).exists():
            return True
        if user_module_root_dir.joinpath(module_name).exists():
            return True
        return False 

    def _load_configurations(self):
        ## reparametrization
        home_dir=Path(self.app['application_dir'])
        ## Function begins
        config_filename=home_dir.joinpath("config.yml")
        environment_filename=home_dir.joinpath("environment.yml")
        config=yaml.safe_load(open(config_filename,'r'))
        environment=yaml.safe_load(open(environment_filename,'r'))
        return config,environment


#### working (API functions)

    def getProtocolTemplateConfig(self):
        # config_dir = self.getConfigDirectory();
        import dtiplayground.dmri.preprocessing.templates as t
        ptc_fn = Path(t.__file__).parent.joinpath('protocol_template.yml')
        # ptc_fn = config_dir.joinpath('protocol_template.yml')
        with open(ptc_fn,'r') as f:
            ptc = yaml.safe_load(f)

        ptc['ui'] = {
            'execution': self.convertTemplate(ptc['options']['io'])
        } 
        return ptc

    def getModuleList(self,extra_dirs = []):
        res = {
            'system': self.getSystemModuleList(),
            'user': self.getUserModuleList(extra_dirs)
        }
        return res

    def getSystemModuleList(self):
        system_dir = Path(self.getSystemModulePath())    
        sp = [x for x in system_dir.glob('*')]
        sp = list(map(lambda x: {'name': x.name, 'path': str(x)},filter(lambda x: x.is_dir() and x.joinpath('__init__.py').exists() , sp)))
        sp.sort(key=lambda x: x['name'])
        return sp

    def getUserModuleList(self,extra_dirs=[]):
        user_module_dir = Path(self.getUserModuleDirectory())
        res = []
        for d in ([user_module_dir] + extra_dirs):      
            up = [x for x in Path(d).glob('*')]
            up = list(map(lambda x: {'name': x.name, 'path': str(x)},filter(lambda x: x.is_dir() and x.joinpath('__init__.py').exists() , up)))
            res = res + up
        res.sort(key=lambda x: x['name'])
        return res

    def getTemplate(self, name, extra_dirs=[]):

        system_dir = Path(self.getSystemModulePath())
        user_module_dir = Path(self.getUserModuleDirectory())
        filepath = system_dir.joinpath(name).joinpath("{}.yml".format(name))
        if not filepath.exists():
            for d in ([user_module_dir] + extra_dirs):
                filepath = d.joinpath(name).joinpath("{}.yml".format(name))
                if filepath.exists(): break
        if not filepath.exists() : raise Exception("There is no such module : {}".format(name))
        with open(filepath,'r') as f:
            original=yaml.safe_load(f)
        ui = self.convertTemplate(original['protocol'])

        protoTemplate = self.getProtocolTemplateConfig()
        
        options = self.convertTemplate(protoTemplate['options']['execution']['options'])
        res = {
                'original': original,
                'ui': {
                    'name' : original['name'],
                    'description': original['description'],
                    'protocol' : self.parseDefaultValues(ui),
                    'options' : options
                 }
        }
        return res


    def setEnvironmentVars(self):
        os.environ['CONFIG_DIR'] = self.getConfigDirectory().__str__()

    def parseDefaultValues(self,template):
        self.setEnvironmentVars()
        for idx,v in enumerate(template):
            print(v['name'])
            if 'filepath' in v['type'] or 'dirpath' in v['type']:
                if v['default_value'] is not None:
                    try:
                        template[idx]['default_value'] = os.path.expandvars(v['default_value'])
                    except:
                        pass

        return template
    def convertTemplate(self, template):
        new_proto = []
        for k,v in template.items():
            tmp = v
            tmp.update({'name': k})
            new_proto.append(tmp)
        return new_proto


    def getConfigDirectory(self):
        return self.app['application_dir']

    def getUserModuleDirectory(self):
        return Path(self.app['config_dir']).joinpath('modules/dmriprep').__str__()

    def getSystemModulePath(self):
        import dtiplayground.dmri.preprocessing.modules as modules
        return Path(modules.__file__).parent.__str__()


##################################################################################
### Module : dmriatlasbuilder.py
### Description : Atlasbuilder api
###
###
###
### Written by : scalphunter@gmail.com ,  2022/10/08
### Copyrights reserved by NIRAL
##################################################################################

import sys, traceback
import os

### API
from flask import request, Response , send_from_directory, jsonify
import json
from . import utils
from pathlib import Path
import yaml
import multiprocessing
from multiprocessing import Process
class DMRIPrepAPI:
    def __init__(self,server,**kwargs):
        self.server = server
        self.app=self.server.app
        self.initEndpoints()

##### Endpoints

    def initEndpoints(self):

        @self.app.route('/api/v1/dmriprep',methods=['GET'])
        def _get_app_info():
            sc=200
            res=None
            req=None
            request_id=utils.get_request_id()
            try:
                res= self.getAppInfo(extra_dirs=[])
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp  

        @self.app.route('/api/v1/dmriprep/generate-default-protocols',methods=['POST'])
        def _post_dmriprep_generate_protocols():
            sc=200
            res=None
            req=None
            request_id=utils.get_request_id()
            try:
                req=request.get_json()
                res= self.generate_default_protocols(req)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp  

        @self.app.route('/api/v1/dmriprep',methods=['POST'])
        def _execute_dmriprep():
            sc=200
            res=None
            req=None
            request_id=utils.get_request_id()
            try:
                req=request.get_json()
                res= self.run(req)
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp  

        @self.app.route('/api/v1/dmriprep/template',methods=['GET'])
        def _get_module_template():
            sc=200
            res=None
            req=None
            request_id=utils.get_request_id()
            param_module_name = request.args.get('name')
            param_dirs_name = request.args.get('dirs', default="[]")

            try:
                res= self.getTemplate(param_module_name, json.loads(param_dirs_name))
                res= utils.add_request_id(res)
            except Exception as e:
                sc=500
                exc=traceback.format_exc()
                res=utils.error_message("{}\n{}".format(str(e),exc),500,request_id)
            finally:
                resp=Response(json.dumps(res),status=sc)
                resp.headers['Content-Type']='application/json'
                return resp  


    def getAppInfo(self, extra_dirs = []):
        from dtiplayground.config import INFO
        version = INFO['dmriprep']['version']

        res = {
            'application' : 'dmriprep',
            'version': version,
            'modules': self.getModuleList(extra_dirs),
            'protocol_template': self.getProtocolTemplateConfig()
        }

        return res

    def generate_default_protocols(self,req):
        from dtiplayground.config import INFO
        req.setdefault('version', INFO['dmriprep']['version'])

        params = {
            'version' : req['version'],
            'output_dir' : req['output_dir'],
            'pipeline' : req['pipeline'],
            'io' : req['io']
        }

        protocols = {
            'version': params['version'],
            'io': params['io'],
            'pipeline': params['pipeline']
        }


        output_dir=Path(params['output_dir'])
        output_dir.mkdir(exist_ok=True,parents=False)

        protocol_fn_path = output_dir.joinpath('protocols.yml')
        protocol_fn_json = output_dir.joinpath('protocols.json')
        yaml.safe_dump(protocols, open(protocol_fn_path,'w'))
        json.dump(protocols, open(protocol_fn_json,'w'),indent=4)
        return req

    # def run(self, req):

    #     params = {
    #         'output_dir' : req['output_dir'],
    #     }
    #     output_dir = Path(params['output_dir'])
    #     config_dir = self.getConfigDirectory()
    #     protocol_fn = output_dir.joinpath('protocols.yml')
    #     protocol = yaml.safe_load(open(protocol_fn,'r'))
    #     params.setdefault('config_dir',str(config_dir))
    #     params.setdefault('protocol_path',str(protocol_fn))
    #     params.setdefault('execution_id', utils.get_uuid())
    #     params.setdefault('global_variables', {})

    
    #     def dmriprep_proc(param):

            
    #         with open(output_dir.joinpath('log.txt'),'w') as sys.stdout:
                
    #             ### begin
    #             os.environ['OMP_NUM_THREADS']=str(protocol['io']['num_threads']) ## this should go before loading any dipy function. 
    #             os.environ['ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS'] = str(protocol['io']['num_threads']) ## for ANTS threading
    #             import  dtiplayground.dmri.common as common
    #             import dtiplayground.dmri.preprocessing
    #             import dtiplayground.dmri.preprocessing.modules as m
    #             import dtiplayground.dmri.preprocessing.protocols as p
    #             logger = common.logger
    #             logger.setFilePointer(sys.stdout)
    #             sys.path.append(Path(m.__file__).parent.__str__()) 
    #             image_list = []
    #             image_list.append(protocol['io']['input_image_1'])
    #             if 'input_image_2' in protocol['io']:
    #                 if protocol['io']['input_image_2'] and protocol['io']['input_image_2'].strip()!='':
    #                     image_list.append(protocol['io']['input_image_2'])

    #             proto= p.Protocols(config_dir,logger=logger)
    #             proto.loadImages(image_list, protocol['io']['baseline_threshold'])
    #             proto.setOutputDirectory(params['output_dir'])
    #             proto.loadProtocols(str(protocol_fn))
    #             proto.setNumThreads(int(protocol['io']['num_threads']))
    #             proto.runPipeline(options=params)
             

    #     res = { 
    #         'execution_id' : params['execution_id'],
    #         'output_dir' : output_dir.__str__()
    #     }
    #     proc = Process(target= dmriprep_proc, name=params['execution_id'],args=[params])
    #     proc.start()

    #     res['pid']=proc.pid
    #     res['proc_name']=proc.name
    #     res['status']='running'
    #     json.dump(res,open(output_dir.joinpath('status.json'),'w'),indent=4)
    #     proc.join()
    #     if proc.exitcode != 0 : 
    #         res['status']='failed'
    #         json.dump(res,open(output_dir.joinpath('status.json'),'w'),indent=4)
    #         raise Exception("Error during running")
    #     else:
    #         res['status']='success'
    #         json.dump(res,open(output_dir.joinpath('status.json'),'w'),indent=4)
    #     return res

    def run(self, req):

        params = {
            'output_dir' : req['output_dir'],
        }
        output_dir = Path(params['output_dir'])
        config_dir = self.getConfigDirectory()
        protocol_fn = output_dir.joinpath('protocols.yml')
        protocol = yaml.safe_load(open(protocol_fn,'r'))
        params.setdefault('config_dir',str(config_dir))
        params.setdefault('protocol_path',str(protocol_fn))
        params.setdefault('execution_id', utils.get_uuid())
        params.setdefault('global_variables', {})

    
        def dmriprep_proc(param):

            
            with open(output_dir.joinpath('log.txt'),'w') as sys.stdout:
                
                ### begin
                os.environ['OMP_NUM_THREADS']=str(protocol['io']['num_threads']) ## this should go before loading any dipy function. 
                os.environ['ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS'] = str(protocol['io']['num_threads']) ## for ANTS threading
                import  dtiplayground.dmri.common as common
                import dtiplayground.dmri.preprocessing
                import dtiplayground.dmri.preprocessing.modules as m
                import dtiplayground.dmri.preprocessing.protocols as p
                from dtiplayground.dmri.preprocessing.app import DMRIPrepApp
                
                inputs = [protocol['io']['input_image_1']]
                protocol['io'].setdefault('input_image_2',None)
                if protocol['io']['input_image_2'] is not None:
                    inputs.append(protocol['io']['input_image_2'])
                options={
                    "input_image_paths" : inputs,
                    "protocol_path" : str(protocol_fn),
                    "output_dir" : protocol['io']['output_directory'],
                    "num_threads":  protocol['io']['num_threads'],
                    "default_protocols": None,
                    "execution_id": param['execution_id'],
                    "baseline_threshold" : protocol['io']['baseline_threshold'],
                    "output_format" : protocol['io']['output_format'],
                    "output_file_base" : protocol['io']['output_filename_base'],
                    "no_output_image" :  protocol['io']['no_output_image'],
                    "global_variables" : {}
                }
                app=DMRIPrepApp(config_root=str(self.server.config_dir))
                app.run(options)

        res = { 
            'execution_id' : params['execution_id'],
            'output_dir' : output_dir.__str__()
        }
        proc = Process(target= dmriprep_proc, name=params['execution_id'],args=[params])
        proc.start()

        res['pid']=proc.pid
        res['proc_name']=proc.name
        res['status']='running'
        json.dump(res,open(output_dir.joinpath('status.json'),'w'),indent=4)
        proc.join()
        if proc.exitcode != 0 : 
            res['status']='failed'
            json.dump(res,open(output_dir.joinpath('status.json'),'w'),indent=4)
            raise Exception("Error during running")
        else:
            res['status']='success'
            json.dump(res,open(output_dir.joinpath('status.json'),'w'),indent=4)
        return res

    def getProtocolTemplateConfig(self):
        # config_dir = self.getConfigDirectory();
        import dtiplayground.dmri.preprocessing.templates as t
        ptc_fn = Path(t.__file__).parent.joinpath('protocol_template.yml')
        # ptc_fn = config_dir.joinpath('protocol_template.yml')
        with open(ptc_fn,'r') as f:
            ptc = yaml.safe_load(f)

        ptc['ui'] = {
            'execution': self.convertTemplate(ptc['options']['io'])
        } 
        return ptc

    def getModuleList(self,extra_dirs = []):
        res = {
            'system': self.getSystemModuleList(),
            'user': self.getUserModuleList(extra_dirs)
        }
        return res

    def getSystemModuleList(self):
        system_dir = Path(self.getSystemModulePath())    
        sp = [x for x in system_dir.glob('*')]
        sp = list(map(lambda x: {'name': x.name, 'path': str(x)},filter(lambda x: x.is_dir() and x.joinpath('__init__.py').exists() , sp)))
        sp.sort(key=lambda x: x['name'])
        return sp

    def getUserModuleList(self,extra_dirs=[]):
        user_module_dir = Path(self.getUserModuleDirectory())
        res = []
        for d in ([user_module_dir] + extra_dirs):      
            up = [x for x in Path(d).glob('*')]
            up = list(map(lambda x: {'name': x.name, 'path': str(x)},filter(lambda x: x.is_dir() and x.joinpath('__init__.py').exists() , up)))
            res = res + up
        res.sort(key=lambda x: x['name'])
        return res

    def getTemplate(self, name, extra_dirs=[]):

        system_dir = Path(self.getSystemModulePath())
        user_module_dir = Path(self.getUserModuleDirectory())
        filepath = system_dir.joinpath(name).joinpath("{}.yml".format(name))
        if not filepath.exists():
            for d in ([user_module_dir] + extra_dirs):
                filepath = d.joinpath(name).joinpath("{}.yml".format(name))
                if filepath.exists(): break
        if not filepath.exists() : raise Exception("There is no such module : {}".format(name))
        with open(filepath,'r') as f:
            original=yaml.safe_load(f)
        ui = self.convertTemplate(original['protocol'])

        protoTemplate = self.getProtocolTemplateConfig()
        
        options = self.convertTemplate(protoTemplate['options']['execution']['options'])
        res = {
                'original': original,
                'ui': {
                    'name' : original['name'],
                    'description': original['description'],
                    'protocol' : self.parseDefaultValues(ui),
                    'options' : options
                 }
        }
        return res

    def defaultVariables(self):
        config_dir = self.getConfigDirectory().__str__()

        resMap = {
            '$CONFIG' : config_dir
        }

        return resMap

    def setEnvironmentVars(self):
        os.environ['CONFIG_DIR'] = self.getConfigDirectory().__str__()

    def parseDefaultValues(self,template):
        # default_map = self.defaultVariables()
        self.setEnvironmentVars()
        for idx,v in enumerate(template):
            if 'filepath' in v['type'] or 'dirpath' in v['type']:
                if v['default_value'] is not None:
                    try:
                        template[idx]['default_value'] = os.path.expandvars(v['default_value'])
                    except:
                        pass

        return template
    def convertTemplate(self, template):
        new_proto = []
        for k,v in template.items():
            tmp = v
            tmp.update({'name': k})
            new_proto.append(tmp)
        return new_proto


    def getConfigDirectory(self):
        from dtiplayground.config import INFO
        version = INFO['dmriprep']['version']
        return Path(os.path.expandvars('$HOME')).joinpath('.niral-dti/dmriprep-{}'.format(version));

    def getUserModuleDirectory(self):
        return Path(os.path.expandvars('$HOME')).joinpath('.niral-dti/modules/dmriprep').__str__()

    def getSystemModulePath(self):
        import dtiplayground.dmri.preprocessing.modules as modules
        return Path(modules.__file__).parent.__str__()



import matplotlib.pyplot as plt
import numpy
import os
import SimpleITK as sitk
import yaml
import copy

from dipy.tracking.benchmarks.bench_streamline import length
from dipy.core.gradients import gradient_table
from dipy.data import load_nifti, get_sphere, default_sphere
from dipy.direction import peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_vtk
from dipy.reconst import dti
from dipy.reconst.shm import CsaOdfModel, OpdtModel
from dipy.segment.mask import median_otsu
from dipy.tracking import utils
from dipy.tracking.local_tracking import LocalTracking 
from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.io.vtk import save_vtk_streamlines, load_vtk_streamlines

from pathlib import Path

import dtiplayground.dmri.preprocessing as prep
import dtiplayground.dmri.common as common
import dtiplayground.dmri.common.tools as tools 
from dtiplayground.dmri.common.dwi import DWI

color = common.Color

class BRAIN_Tractography(prep.modules.DTIPrepModule):
    def __init__(self,config_dir,*args,**kwargs):
        super().__init__(config_dir,*args,**kwargs)
        global logger
        logger = self.logger.write

    def generateDefaultProtocol(self,image_obj):
        super().generateDefaultProtocol(image_obj)
        ## todos
        return self.protocol

    def process(self,*args,**kwargs): ## variables : self.config_dir, self.source_image, self.image (output) , self.result_history , self.result (output) , self.protocol, self.template
        super().process()
        inputParams=self.getPreviousResult()['output']
        # << TODOS>>
        # create nifti files in output_dir

## old
        input_dwi = Path(self.output_dir).joinpath('input.nii.gz').__str__()
        self.writeImageWithOriginalSpace(input_dwi,'nifti',dtype='float')
        input_bval = input_dwi.replace('.nii.gz', '.bval')
        input_bvec = input_dwi.replace('.nii.gz', '.bvec')

        import nibabel as nib
        img = nib.load(input_dwi)
        # # load files
        # data, affine, img = load_nifti(input_dwi, return_img=True)
        # bvals, bvecs = read_bvals_bvecs(input_bval, input_bvec)
        # gradient_tab = gradient_table(bvals, bvecs)
## old ends
## new code
        data = self.image.images
        affine = self.image.getAffineMatrixForNifti()
        # affine=img.affine
        bvecs = [x['nifti_gradient'] for x in self.image.getGradients()]
        bvals = [x['b_value'] for x in self.image.getGradients()]
        gradient_tab = gradient_table(bvals,bvecs)

## new code ends

        # get brain mask
        masked_data, brainmask = median_otsu(data, vol_idx=[0], numpass=1)
        dilated_mask=None
        if self.protocol['referenceTractFile'] is not None:
            logger("Partial tractography mode",color.INFO)
            dilated_mask = self.RegisterSingleTract(**self.protocol)
            # brainmask = brainmask * dilated_mask
        else:
            logger("No reference tracts are set, whole brain tractography mode is set",color.INFO)
        # get FA
        
        dti_model = dti.TensorModel(gradient_tab)
        dti_fit = dti_model.fit(masked_data, mask=brainmask)
        fa = dti_fit.fa

        # saving tensor file to nrrd
        logger("Saving tensorfile..",color.PROCESS)
        self.saveTensor(dti_fit)
        logger("Tensor Saved",color.OK)

        # get WM mask
        if self.protocol['whiteMatterMaskThreshold'] == 'manual':
            WM_mask = self.GetWMMaskManualThreshold(fa)
        #if self.protocol['whiteMatterMaskThreshold'] == 'otsu':
        #    WM_mask = self.GetWMMaskOtsu(fa)
            if dilated_mask is not None:
                WM_mask = WM_mask * dilated_mask


        fig = plt.figure()
        plt.xticks([])
        plt.yticks([])
        fig.tight_layout()
        plt.imshow(WM_mask[:, :, data.shape[2] // 2].T, cmap='gray',
                   origin='lower', interpolation='nearest')
        fig.savefig(Path(self.output_dir).joinpath('white_matter_mask.png').__str__())
        
        # generate peaks
        if self.protocol['method'] == 'tensor':
            peaks = self.GeneratePeaksTensor(masked_data, dti_model, mask=None)
        elif self.protocol['method'] == 'csa':
            peaks = self.GeneratePeaksCSA(gradient_tab, masked_data, mask=None)
        elif self.protocol['method'] == 'opdt':
            peaks = self.GeneratePeaksOPDT(gradient_tab, masked_data, mask=None)

        logger("Method: {} was selected".format(self.protocol['method']),color.INFO)
        # generate seeds
        seeds = utils.seeds_from_mask(WM_mask, affine, density=[1, 1, 1])

        # generate tracts
        logger("Generating streamlines ...",color.PROCESS)
        stopping_criterion = ThresholdStoppingCriterion(fa, self.protocol['stoppingCriterionThreshold'])
        streamlines_generator = LocalTracking(peaks, stopping_criterion, seeds, affine=affine, step_size=.3)
        streamlines = Streamlines(streamlines_generator)
        logger("Streamlines generated.",color.OK)
        # filter tracts
        if self.protocol['removeShortTracts'] == True:
            streamlines = self.RemoveShortTracts(streamlines, self.protocol['shortTractsThreshold'])
        if self.protocol['removeLongTracts'] == True:
            streamlines = self.RemoveLongTracts(streamlines, self.protocol['longTractsThreshold'])

        # save tracts
        # sft = StatefulTractogram(streamlines, img, Space.RASMM)
        # sft = StatefulTractogram(streamlines, img, Space.VOXMM)
        tract_path = Path(self.output_dir).joinpath('tractogram.vtk').__str__()
        tract_path_vtp = Path(self.output_dir).joinpath('tractogram.vtp').__str__()
        # save_vtk(sft, tract_path, bbox_valid_check=False)
        save_vtk_streamlines(streamlines, tract_path, to_lps=False, binary=False)
        save_vtk_streamlines(streamlines, tract_path_vtp, to_lps=False, binary=False)
        self.addOutputFile(tract_path, "tractogram")
        self.addOutputFile(tract_path_vtp, "tractogram")
        self.result['output']['success']=True
        return self.result

    @common.measure_time
    def saveTensor(self, fitted):
        ## convert 3x3 symmetric matrices to xx,xy,xz,yy,yz,zz vectors
        np=numpy
        logger("Reducing 3x3 symmetric matrix to vector")
        def uppertriangle(matrix):
            outvec=[]
            for i in range(3):
                for j in range(i,3):
                    outvec.append(matrix[i,j])
            return np.array(outvec)
        quad_form = fitted.quadratic_form
        new_quadform = np.ndarray(shape=(quad_form.shape[0],quad_form.shape[1],quad_form.shape[2],6),dtype=float)
        for d1 in range(quad_form.shape[0]):
            for d2 in range(quad_form.shape[1]):
                for d3 in range(quad_form.shape[2]):
                    mat = quad_form[d1,d2,d3]
                    new_quadform[d1,d2,d3]=uppertriangle(mat)

        # TODO : make nrrd file for new_quadform image volume (kind will be "3D-symmetric-matrix") , ref: http://teem.sourceforge.net/nrrd/format.html
        temp_dti_image = DWI()
        temp_dti_image.copyFrom(self.image, image=False, gradients=False)
        temp_dti_image.setImage(new_quadform,modality='DTI', kinds=['space','space','space','3D-symmetric-matrix'])
        dti_filename=Path(self.output_dir).joinpath('tensor.nrrd').__str__()
        sp_dir=self.getSourceImageInformation()['space']
        temp_dti_image.setSpaceDirection(target_space=sp_dir)
        temp_dti_image.writeImage(dti_filename,dest_type='nrrd',dtype="float32")
        self.addOutputFile(dti_filename, 'DTI')
        self.addGlobalVariable('dti_path',dti_filename)

    @common.measure_time
    def GetWMMaskManualThreshold(self, fa):
        WM1 = fa > self.protocol['thresholdLow']
        WM2 = fa < self.protocol['thresholdUp']
        WM_mask_array = numpy.logical_and(WM1, WM2)
        WM_mask_array = WM_mask_array.astype(int)
        WM_mask_image = sitk.GetImageFromArray(WM_mask_array)
        WM_mask_array = self.MorphologicalOpeningWMMask(WM_mask_image)
        return WM_mask_array

    @common.measure_time
    def GetWMMaskOtsu(self, fa):
        otsu_filter = sitk.OtsuMultipleThresholdsImageFilter()
        otsu_filter.SetNumberOfThresholds(3)
        fa_image = sitk.GetImageFromArray(fa)
        WM_mask_image = otsu_filter.Execute(fa_image)
        (t1, t2, t3)= otsu_filter.GetThresholds()
        WM_mask_array = self.MorphologicalOpeningWMMask(WM_mask_image)
        WM_mask_array = WM_mask_array >= 3
        return WM_mask_array

    @common.measure_time
    def MorphologicalOpeningWMMask(self, WM_mask_image):
        opening = sitk.BinaryMorphologicalOpeningImageFilter()
        opening.SetKernelType(sitk.sitkCross)
        opening.SetKernelRadius(1)
        WM_mask_image = opening.Execute(WM_mask_image)
        WM_mask_array = sitk.GetArrayFromImage(WM_mask_image)
        return WM_mask_array

    @common.measure_time
    def GeneratePeaksTensor(self, masked_data, dti_model,mask=None):
        sphere = get_sphere('symmetric362')
        peaks = peaks_from_model(model=dti_model,
            data=masked_data,
            sphere=sphere,
            relative_peak_threshold=self.protocol['relativePeakThreshold'],
            min_separation_angle=self.protocol['minPeakSeparationAngle'],
            mask=mask,
            npeaks=1)
        return peaks

    @common.measure_time
    def GeneratePeaksCSA(self, gtab, masked_data, mask=None):
        csa_model = CsaOdfModel(gtab, sh_order=2)#self.protocol['shOrder'])
        peaks = peaks_from_model(model=csa_model,
            data=masked_data,
            sphere=default_sphere,
            relative_peak_threshold=self.protocol['relativePeakThreshold'],
            min_separation_angle=self.protocol['minPeakSeparationAngle'],
            mask=mask)
        return peaks

    @common.measure_time
    def GeneratePeaksOPDT(self, gtab, masked_data, mask=None):
        opdt_model = OpdtModel(gtab, sh_order=2)#self.protocol['shOrder'])
        peaks = peaks_from_model(opdt_model, 
            data=masked_data,
            sphere=default_sphere,
            relative_peak_threshold=self.protocol['relativePeakThreshold'],
            min_separation_angle=self.protocol['minPeakSeparationAngle'],
            mask=mask)
        return peaks

    @common.measure_time
    def RemoveShortTracts(self, streamlines, threshold):
        streamlines_length = length(streamlines)
        number_of_streamlines = len(streamlines_length)
        long_streamlines = [streamlines[i] for i in range(number_of_streamlines) if streamlines_length[i] > threshold]
        return long_streamlines
    
    @common.measure_time
    def RemoveLongTracts(self, streamlines, threshold):
        streamlines_length = length(streamlines)
        number_of_streamlines = len(streamlines_length)
        short_streamlines = [streamlines[i] for i in range(number_of_streamlines) if streamlines_length[i] < threshold]
        return short_streamlines

## single tract registration

    @common.measure_time
    def RegisterSingleTract(self, **protocol):
        res = {}
        inputDTI = Path(self.output_dir).joinpath('input.nrrd').__str__()
        if 'dti_path' in self.global_variables:
            inputDTI = self.global_variables['dti_path']
        else:
            self.writeImageWithOriginalSpace(inputDTI,'nrrd',dtype='float')
        inputFiberFile = protocol['referenceTractFile']
        displacementField = protocol['displacementFieldFile']
        if not displacementField:
            displacementField = self.global_variables['displacement_field_path']
            if not Path(self.global_variables['displacement_field_path']).exists():
                raise Exception("Couldn't find the displacement field file")

        outputFiberTract = Path(self.output_dir).joinpath('registered_ref_tract.vtk').__str__()
        
    # Register reference tract with the displacement field 
        niralutils = tools.NIRALUtilities(softwares=self.softwares)
        niralutils.dev_mode=True
        arguments = ['--polydata_input', inputFiberFile,
                     '-o', outputFiberTract,
                     '-D', displacementField,
                     '--inverty',
                     '--invertx']
        if self.overwriteFile(outputFiberTract) : output=niralutils.polydatatransform(arguments)
        self.addGlobalVariable('reference_tract_path', outputFiberTract) ## update tract path to the updated one

    ## Dilation and voxelization of the mapped reference tracts , getting labelmap
        labelMapFile = Path(self.output_dir).joinpath('labelmap.nrrd').__str__()
        arguments = ['--voxelize', labelMapFile,
                     '--fiber_file', outputFiberTract,
                     '-T', inputDTI]
        fiberprocess = tools.FiberProcess(softwares=self.softwares)
        fiberprocess.dev_mode = True
        if self.overwriteFile(labelMapFile) : fiberprocess.execute(arguments=arguments)
        self.addGlobalVariable('labelmap_path', labelMapFile)

    ## Dilation and voxelization of the reference tracts
        dilatedLabelmapFile = Path(self.output_dir).joinpath('labelmap_dilated.nrrd').__str__()
        dilationRadius = self.protocol['dilationRadius']
        imagemath = tools.ImageMath(softwares=self.softwares)
        imagemath.dev_mode=True
        arguments = [labelMapFile,
                     '-dilate', str(dilationRadius)+',1',
                     '-outfile', dilatedLabelmapFile
                     ]
        if self.overwriteFile(dilatedLabelmapFile) : imagemath.execute(arguments=arguments)
        self.addGlobalVariable('labelmap_path', dilatedLabelmapFile)
        labelMapImage = common.dwi.DWI(labelMapFile)
        dilatedLabelmapImage = common.dwi.DWI(dilatedLabelmapFile)
        dipyLabelMap = labelMapImage.images + dilatedLabelmapImage.images

        return dipyLabelMap





#
#   utils.py 
#   2022-10-09
#   Written by SK Park, NIRAL, UNC
#
#   Atlasbuilding utilities
#


import os # To run a shell command : os.system("[shell command]")
import sys # to return an exit code
import shutil # to remove a non empty directory
import json
import argparse
import csv 
import copy 
import xml.etree.cElementTree as ET 
import xml.dom.minidom as minidom
from pathlib import Path 

import dtiplayground.dmri.atlasbuilder as ab 
import dtiplayground.dmri.common.tools as tools
import dtiplayground.dmri.common as common

logger=common.logger.write

def dumpXml(xml):
    root = xml.getroot()
    xmlstr = minidom.parseString(ET.tostring(root)).toprettyxml(indent="\t")
    return xmlstr

def makeXmlElementTree(config, greedy): ## generate greedy xml from the scratch (recommended)
    root = ET.Element('ParameterFile')
    wis = ET.SubElement(root,'WeightedImageSet')
    ### WeightedImageSet
    siw = ET.SubElement(wis, 'ScaleImageWeights', { 'val':'true' })
    for cid in config['m_CasesIDs']:
        wi=ET.Element('WeightedImage',{})
        lastLoop=str(config['m_nbLoops'])
        p=os.path.join(config['m_OutputPath'],"1_Affine_Registration/Loop"+lastLoop+"/"+cid+"_Loop"+lastLoop+"_Final"+config["m_ScalarMeasurement"]+".nrrd")
        wiFilename=ET.Element('Filename',{'val':str(p)})
        wiItkTransform=ET.Element('ItkTransform',{'val':'1'})
        wi.insert(0,wiFilename)
        wi.insert(1,wiItkTransform)
        wis.insert(-1, wi)

    iif = ET.SubElement(wis, 'InputImageFormatString')
    fs = ET.SubElement(iif, 'FormatString', { 'val':'' })
    base = ET.SubElement(iif, 'Base', {'val':'0' })
    num_files = ET.SubElement(iif, 'NumFiles', { 'val' : str(len(config['m_CasesIDs'])) } )
    
    weight = ET.SubElement(iif, 'Weight', { 'val': '1' })

    ### GreedyScaleLevel
    for row in greedy['rows']:
        gsl = ET.SubElement(root,'GreedyScaleLevel',{})
        scale_level = ET.SubElement(gsl, 'ScaleLevel')
        downsample_factor = ET.SubElement(scale_level, 'DownSampleFactor')
        downsample_factor.set('val',str(row['scale_level']))
        n_iteration = ET.SubElement(gsl, 'NIterations')
        n_iteration.set('val',str(row['n_iterations']))
        iterator = ET.SubElement(gsl,'Iterator')
        maxpert = ET.SubElement(iterator,'MaxPert')
        maxpert.set('val',str(row['max_perturbation']))
        diffoper = ET.SubElement(iterator, 'DiffOper')
        alpha = ET.SubElement(diffoper,'Alpha')
        alpha.set('val',str(row['alpha']))
        beta = ET.SubElement(diffoper,'Beta')
        beta.set('val', str(row['beta']))
        gamma = ET.SubElement(diffoper,'Gamma')
        gamma.set('val', str(row['gamma']))
        
    ### Other options
    n_threads = ET.SubElement(root, 'nThreads')
    n_threads.set('val', str(config['m_NbThreadsString']))
    outputprefix = ET.SubElement(root, 'OutputPrefix')
    outputprefix.set('val', config["m_OutputPath"]+"/2_NonLinear_Registration/")
    outputsuffix = ET.SubElement(root, 'OutputSuffix')
    outputsuffix.set('val', 'mhd')
    xml = ET.ElementTree(element=root)
    return xml

# def makeXmlElementTreeDefault(cfg): ## generate them by default template (not recommended)
#     xmlfile=cfg["m_GreedyAtlasParametersTemplatePath"]
#     x=ET.parse(xmlfile)
#     r=x.getroot()

#     ## remove all dummy dataset files
#     wis=r.find('WeightedImageSet')
#     wi_list=wis.findall('WeightedImage')
#     for w in wi_list:
#         wis.remove(w)

#     ## insert new dataset
#     for cid in cfg["m_CasesIDs"]:
#         wi=ET.Element('WeightedImage',{})
#         lastLoop=str(cfg['m_nbLoops'])
#         p=os.path.join(cfg['m_OutputPath'],"1_Affine_Registration/Loop"+lastLoop+"/"+cid+"_Loop"+lastLoop+"_Final"+cfg["m_ScalarMeasurement"]+".nrrd")
#         wiFilename=ET.Element('Filename',{'val':str(p)})
#         wiItkTransform=ET.Element('ItkTransform',{'val':'1'})
#         wi.insert(0,wiFilename)
#         wi.insert(1,wiItkTransform)
#         wis.insert(-1,wi)  ## insert to the last

#     ## change output path 
#     for neighbor in r.iter('OutputPrefix'):
#         logger("{} {}".format(neighbor.tag,neighbor.attrib))
#         neighbor.set('val',cfg["m_OutputPath"]+"/2_NonLinear_Registration/")

#     return x

def generateGreedyAtlasParametersFile(cfg, greedy):

    # if greedy_scale_table is not None:
    x = makeXmlElementTree(cfg, greedy)
    # else:
    #     x = makeXmlElementTreeDefault(cfg)

    outputfile=cfg["m_OutputPath"]+"/2_NonLinear_Registration/GreedyAtlasParameters.xml"
    xmlstr = dumpXml(x)
    # x.write(outputfile)
    with open(outputfile, 'w') as f:
        f.writelines(xmlstr)



def DisplayErrorAndQuit ( Error ):
    msg='\n\nERROR DETECTED IN WORKFLOW:'+Error
    logger(msg)
    logger('ABORT')
    raise Exception(msg)


# Function that checks if file exist and replace old names by new names if needed
def CheckFileExists ( File, case, caseID ) : # returns 1 if file exists or has been renamed and 0 if not
    if os.path.isfile( File ) : # file exists
      return 1
    else : # file does not exist: check if older version of file can exist (if file name has been changed)
      NamesThatHaveChanged = ["MeanImage", "DiffeomorphicDTI", "DiffeomorphicAtlasDTI", "HField", "GlobalDisplacementField"] # latest versions of the names
      if any( changedname in File for changedname in NamesThatHaveChanged ) : # if name has been changed, check if older version files exist
        if "MeanImage" in File :
          OldFile = File.replace("Mean", "Average")
          if os.path.isfile( OldFile ) : # old file exists: rename and return 1
            os.rename(OldFile, File)
            return 1
          else:
            return 0
        if "DiffeomorphicDTI" in File :
          OldFile = File.replace( caseID, "Case" + str(case+1) ).replace("DiffeomorphicDTI", "AWDTI")
          if os.path.isfile( OldFile ) : # old file exists: rename and return 1
            os.rename(OldFile, File)
            os.rename(OldFile.replace("AWDTI","AW"+config['m_ScalarMeasurement']), File.replace("DiffeomorphicDTI","Diffeomorphic"+config['m_ScalarMeasurement']))
            os.rename(OldFile.replace("AWDTI","AWDTI_float"), File.replace("DiffeomorphicDTI","DiffeomorphicDTI_float"))
            return 1
          else : # test other old name
            OldFile = File.replace( caseID, "Case" + str(case+1) )
            if os.path.isfile( OldFile ) : # old file exists: rename and return 1
              os.rename(OldFile, File)
              os.rename(OldFile.replace("DiffeomorphicDTI","Diffeomorphic"+config['m_ScalarMeasurement']), File.replace("DiffeomorphicDTI","Diffeomorphic"+config['m_ScalarMeasurement']))
              os.rename(OldFile.replace("DiffeomorphicDTI","DiffeomorphicDTI_float"), File.replace("DiffeomorphicDTI","DiffeomorphicDTI_float"))
              return 1
            else:
              return 0
        if "DiffeomorphicAtlasDTI" in File :
          OldFile = File.replace("DiffeomorphicAtlasDTI", "AWAtlasDTI")
          if os.path.isfile( OldFile ) : # old file exists: rename and return 1
            os.rename(OldFile, File)
            os.rename(OldFile.replace("AWAtlasDTI","AWAtlas"+config['m_ScalarMeasurement']), File.replace("DiffeomorphicAtlasDTI","DiffeomorphicAtlas"+config['m_ScalarMeasurement']))
            os.rename(OldFile.replace("AWAtlasDTI","AWAtlasDTI_float"), File.replace("DiffeomorphicAtlasDTI","DiffeomorphicAtlasDTI_float"))
            return 1
          else:
            return 0
        if "HField" in File :
          OldFile = File.replace( caseID, "Case" + str(case+1) ).replace("H", "Deformation")
          if os.path.isfile( OldFile ) : # old file exists: rename and return 1
            os.rename(OldFile, File)
            return 1
          else : # test other old name
            OldFile = File.replace( caseID, "Case" + str(case+1) )
            if os.path.isfile( OldFile ) : # old file exists: rename and return 1
              os.rename(OldFile, File)
              return 1
            else:
              return 0
        if "GlobalDisplacementField" in File :
          OldFile = File.replace( caseID, "Case" + str(case+1) ).replace("Displacement", "Deformation")
          if os.path.isfile( OldFile ) : # old file exists: rename and return 1
            os.rename(OldFile, File)
            return 1
          else : # test other old name
            OldFile = File.replace( caseID, "Case" + str(case+1) )
            if os.path.isfile( OldFile ) : # old file exists: rename and return 1
              os.rename(OldFile, File)
              return 1
            else:
              return 0
      else: # file does not exist and name has not been changed: check if the caseX version exists
        if caseID : # CaseID is empty for averages
          OldFile = File.replace( caseID, "Case" + str(case+1) )
          if os.path.isfile( OldFile ) : # old file exists: rename and return 1
            os.rename(OldFile, File)
            return 1
          else:
            return 0
        else: # for averages
          return 0

def _check_deformation_sequence_file(config,payload):
    binaryPath=config["m_SoftPath"][9]
    if(not os.path.exists(binaryPath)):
        logger("Software is missing : %s" %binaryPath)
        return False
    logger("Software path is : %s " % binaryPath)
    res=True
    for s in payload:
        cnt=0
        for f in s["filelist"]:
            if(not os.path.exists(f)):
                logger("Error while checking : %s " % f)
                cnt+=1
                res=False
        if cnt==0:
            logger("%s : OK" % s["id"])
        else:
            logger("%s : NOT OK" %s['id'])
    return res 

def _generate_concatenated_displacement_directory(config):
    projectDir=config["m_OutputPath"]
    outDir=os.path.join(projectDir,"displacement_fields")
    if not os.path.isdir(outDir):
      logger("\n=> Creation of the directory containing concatenated displacement fields = " + outDir)
      os.mkdir(outDir)
    return outDir

def ITKTransformTools_Concatenate(config,payload): ## payload should be deformation_track.json list
    command=""
    if(_check_deformation_sequence_file(config,payload)):
        outputDir=_generate_concatenated_displacement_directory(config)
        binaryPath=config["m_SoftPath"][9] 
        command+=binaryPath + " concatenate "
        refDTI=config["m_OutputPath"] + "/final_atlas/5_Final_Atlas/FinalAtlasDTI_float.nrrd"
        for idx, elm in enumerate(payload):
            hpairList=elm["id"].split("/")
            outFilename="_".join(hpairList) + "_GlobalDisplacementField_Concatenated.nrrd"
            outFilename=os.path.join(outputDir,outFilename)
            if Path(outFilename).exists() and (not config['m_Overwrite']==1):
                logger("File : {} already exists.".format(outFilename))
                continue
            #logger("Output filename : %s"%outFilename)
            tmpCommand=command + outFilename +" -r " + refDTI + " "
            inpListStr=""
            fl=map(str,elm["filelist"])
            #fl.reverse()
            for fn in fl:
                inpListStr+= fn + " displacement "
            tmpCommand+=inpListStr
            #logger("%d : %s " %(idx,tmpCommand))
            os.system(tmpCommand)
    else:
        logger("There are some missing deformation fields file(s)")
        raise(Exception("There are some missing deformation fields file(s)"))

def ITKTransformTools_Concatenate_Inverse(config,payload): ## payload should be deformation_track.json list
    command=""
    if(_check_deformation_sequence_file(config,payload)):
        outputDir=_generate_concatenated_displacement_directory(config)
        binaryPath=config["m_SoftPath"][9] 
        command+=binaryPath + " concatenate "
        for idx, elm in enumerate(payload):
            refDTI=elm['original_dti_path']
            hpairList=elm["id"].split("/")
            outFilename=elm['output_path']
            if Path(outFilename).exists() and (not config['m_Overwrite']==1):
                logger("File : {} already exists.".format(outFilename))
                continue
            #logger("Output filename : %s"%outFilename)
            tmpCommand=command + outFilename +" -r " + refDTI + " "
            inpListStr=""
            fl=map(str,elm["filelist"])
            #fl.reverse()
            for fn in fl:
                inpListStr+= fn + " displacement "
            tmpCommand+=inpListStr
            #logger("%d : %s " %(idx,tmpCommand))
            os.system(tmpCommand)
    else:
        logger("There are some missing deformation fields file(s)")
        raise(Exception("There are some missing deformation fields file(s)"))



def unique(list1): 
    unique_list = [] 
    for x in list1: 
        # check if exists in unique_list or not 
        if x not in unique_list: 
            unique_list.append(x) 
    return unique_list

def isComponent(seq,name):
    comp=list(filter(lambda x : x['name']==name,seq))
    if len(comp)>0 :
        return comp[0] 
    else:
        return False 

def find_config_by_nodename(build_sequence,nodename):
    for cfg in build_sequence:
        if cfg["m_NodeName"]==nodename:
            return cfg 


def generate_deformation_track(seq,node="target"): #input : initialSequence to generate deformation field tracking information (to concatenate them)
    component=isComponent(seq,node)
    outseq=[]

    if component != False:
        for c in component["dataset_ids"]:
            tmpseq=generate_deformation_track(seq,c)
            for t in tmpseq:
                outseq.append(node+"/"+t)
    else:
        outseq.append(node)
        return outseq 
    return outseq

def invert_deformation_track(deformation_seq):
    seq=copy.copy(deformation_seq)
    outseq=[]
    for s in seq:
        elm=s
        strvec=s['id'].split("/")
        strvec.reverse()
        elm['id']='/'.join(strvec)
        #elm['original_dti_id']=strvec[-1]
        arr=[]
        for e in s['filelist']:
            basedir=os.path.dirname(e)
            name="_".join(os.path.basename(e).split('_')[:-1])+"_InverseGlobalDisplacementField.nrrd"
            inverted_deform_path=os.path.join(basedir,name)
            arr.append(inverted_deform_path)
        arr.reverse()
        elm['filelist']=arr 

        output_dir=os.path.dirname(s['output_path'])
        output_name="_".join(os.path.basename(s['output_path']).split('_')[:-2])+"_InverseGlobalDisplacementField_Concatenated.nrrd"
        output_path=os.path.join(output_dir,output_name)
        elm['output_path']=output_path
        outseq.append(elm)
    return outseq 

    
def furnish_deformation_track(seq,project_path,build_sequence,inverse=False): #input deformSequence 
    res=[]
    for d in seq:
        tmp={}
        tmp['id']=d
        compseq=d.split('/')
        cfg=find_config_by_nodename(build_sequence,compseq[-2])
        originalDTIId=compseq[-1]
        originalDTIPath=None
        for idx,case in enumerate(zip(cfg["m_CasesIDs"],cfg["m_CasesPath"])):
            caseID,casePath=case 
            if originalDTIId==caseID: 
                originalDTIPath=casePath
                break

        entry=[]
        for idx,c in enumerate(compseq[0:-1]):
            fpath="atlases/" + c + "/5_Final_Atlas/FinalDeformationFields/" + compseq[idx+1] + "_GlobalDisplacementField.nrrd"
            fpath=os.path.join(project_path,fpath)
            entry.append(fpath)
        tmp['filelist']=entry
        tmp['original_dti_path']=originalDTIPath 
        tmp['original_dti_id']=originalDTIId
        tmp['scalar_measurement']=cfg["m_ScalarMeasurement"]
        tmp['nb_loops']=cfg['m_nbLoops']
        tmp['nb_loops_dtireg']=cfg['m_nbLoopsDTIReg']
        tmp['project_path']=cfg['m_OutputPath']
        tmp['need_to_be_cropped']=cfg['m_NeedToBeCropped']
        outputDir=os.path.join(project_path,"displacement_fields")
        hpairList=tmp["id"].split("/")
        outFilename="_".join(hpairList) + "_GlobalDisplacementField_Concatenated.nrrd"
        outFilename=os.path.join(outputDir,outFilename)
        tmp['output_path']=outFilename
        res.append(tmp)
    return res 




def parse_hbuild(hb,root_path,root_node="target"): #hbuild parser to generate build sequence
    if root_node is None:
        root_node=hb['project']['target_node']
    root=hb['build'][root_node]
    seq=[]
    nodeFiles=[] ## sub node's final atlases
    # scalar=hb['config']['m_ScalarMeasurement']
    if root["type"]=="node":    
        for c in root["components"]:
            seq+=parse_hbuild(hb, root_path=root_path, root_node=c)
            nodeAtlasPath=os.path.join(root_path,"atlases/"+c+"/5_Final_Atlas/FinalAtlasDTI_float.nrrd")
            nodeFiles.append(nodeAtlasPath)
    elif root["type"]=="end_node":
        if root["filetype"]=="dataset":
            rows=[]
            rows_id=[]
            with open(str(root['datasetfiles']),'r') as f:
                csvreader=csv.reader(f)
                next(csvreader,None)
                for r in csvreader:
                    fpath=str(r[1])
                    fid=os.path.splitext(os.path.basename(fpath))[0]
                    rows.append(fpath)
                    rows_id.append(str(fid))

            return  [{"name" : str(root_node),
                "dataset_files" : rows,
                "dataset_ids" : rows_id,
                "project_path" : str(os.path.join(root_path,"atlases/"+root_node))
                }]
        else:
            flist=list(map(str,root["datasetfiles"]))
            fids=[]
            for e in flist:
                fid=os.path.splitext(os.path.basename(e))[0]
                fids.append(fid)

            return [{"name" : str(root_node),
                    "dataset_files" : flist,
                    "dataset_ids" : fids ,
                    "project_path" : str(os.path.join(root_path,"atlases/"+root_node))
                    }]

    # node type file reading

    seq+=[{"name" : str(root_node),
            "dataset_files" : list(map(str,nodeFiles)),
            "dataset_ids" : list(map(str,root["components"])),
            "project_path" : str(os.path.join(root_path,"atlases/"+root_node))

         }]
    seq=unique(seq)

    ## generate final buildsequence furnished with configuration


    return seq

def furnish_sequence(hb,seq):
    bs=[]
    for s in seq:
        conf=hb["config"].copy()
        conf["m_OutputPath"]=s['project_path']
        conf["m_CasesPath"]=s['dataset_files']
        conf["m_CasesIDs"]=s['dataset_ids']
        conf["m_NodeInfo"]=hb["build"][s['name']]
        conf["m_NodeName"]=s["name"]
        bs.append(conf)

    return bs

def generate_directories(project_path,sequence): ## from build sequence, generate directories
    atlasesPath=os.path.join(project_path,"atlases")
    finalAtlasPath=os.path.join(project_path,"final_atlas")
    if not os.path.isdir(atlasesPath):
      logger("\n=> Creation of the atlas directory = " + atlasesPath)
      os.mkdir(atlasesPath)
    if not os.path.isdir(finalAtlasPath):
      logger("\n=> Creation of the atlas directory = " + finalAtlasPath)
      os.mkdir(finalAtlasPath)
    for s in sequence:
        apath=os.path.join(s["m_OutputPath"])
        if not os.path.isdir(apath):
          logger("\n=> Creation of the atlas directory = " + apath)
          os.mkdir(apath)
    logger("Initial directories are generated")


def dependency_satisfied(hb,node_name,completed_atlases):
    if hb["build"][node_name]["type"]=="end_node": 
        return True
    else:
        comps=hb["build"][node_name]["components"]
        for c in comps:
            if c not in completed_atlases: return False 
        return True



def generate_results_csv_from_deformation_track(deformation_track,project_path): # generate final result file with deformation track file

    dt=deformation_track
    outpath=os.path.join(project_path,"DTIAtlasBuilderResults.csv")
    
    m_ScalarMeasurement=dt[0]["scalar_measurement"]
    m_NeedToBeCropped=dt[0]["need_to_be_cropped"]
    header=["id", "Original DTI Image"]
    if m_NeedToBeCropped==1: header + ["Cropped DTI"]
    tmp=[m_ScalarMeasurement+ " from original",
        "Affine transform", "Affine Registered DTI", 
        "Affine Registered "+m_ScalarMeasurement,
        "Diffeomorphic Deformed " + m_ScalarMeasurement,
        "Diffeomorphic Deformation field to Affine space",
        "Diffeomorphic Deformation field to Affine space",
        "Diffeomorphic DTI",
        "Diffeomorphic Deformation field to Original space",
        "DTI-Reg Final DTI"
        ]
    header+=tmp
    with open(outpath,"w") as f:
        csvwriter=csv.writer(f,delimiter=',')
        csvwriter.writerow(header)
        for idx,case in enumerate(dt):
            caseID,casePath = case["original_dti_id"],case["original_dti_path"]
            m_OutputPath=case["project_path"]
            m_nbLoops=case["nb_loops"]
            m_nbLoopsDTIReg=case["nb_loops_dtireg"]
            row=[
                idx+1,
                casePath]
            if m_NeedToBeCropped==1: row+=[m_OutputPath+"/1_Affine_Registration/" + caseID+"_croppedDTI.nrrd"]
            concatenated_displacement_path=case["output_path"]
            row+=[
                m_OutputPath+"/1_Affine_Registration/" + caseID + "_" + m_ScalarMeasurement + ".nrrd",
                m_OutputPath+"/1_Affine_Registration/Loop" + str(m_nbLoops) + "/" + caseID + "_Loop" + str(m_nbLoops)+"_LinearTrans.txt",
                m_OutputPath+"/1_Affine_Registration/Loop" + str(m_nbLoops) + "/" + caseID + "_Loop" + str(m_nbLoops)+"_LinearTrans_DTI.nrrd",
                m_OutputPath+"/1_Affine_Registration/Loop" + str(m_nbLoops) + "/" + caseID + "_Loop" + str(m_nbLoops)+"_Final" + m_ScalarMeasurement +".nrrd",
                m_OutputPath+"/2_NonLinear_Registration/" + caseID + "_NonLinearTrans_" + m_ScalarMeasurement + ".mhd",
                m_OutputPath+"/2_NonLinear_Registration/" + caseID + "_HField.mhd" ,
                m_OutputPath+"/2_NonLinear_Registration/" + caseID + "_InverseHField.mhd" ,
                m_OutputPath+"/3_Diffeomorphic_Atlas/" + caseID + "_DiffeomorphicDTI.nrrd",
                concatenated_displacement_path,
                m_OutputPath+"/4_Final_Resampling/FinalTensors/" + caseID + "_FinalDeformedDTI.nrrd"
            ]
            csvwriter.writerow(row)

def generate_results_csv(cfg):

    outpath=os.path.join(cfg["m_OutputPath"],"DTIAtlasBuilderResults.csv")
    m_OutputPath=cfg["m_OutputPath"]
    m_ScalarMeasurement=cfg["m_ScalarMeasurement"]
    m_nbLoops=cfg["m_nbLoops"]
    m_nbLoopsDTIReg=cfg["m_nbLoopsDTIReg"]
    m_NeedToBeCropped=cfg["m_NeedToBeCropped"]
    header=["id", "Original DTI Image"]
    if m_NeedToBeCropped==1: header + ["Cropped DTI"]
    tmp=[cfg["m_ScalarMeasurement"]+ " from original",
        "Affine transform", "Affine Registered DTI", 
        "Affine Registered "+cfg["m_ScalarMeasurement"],
        "Diffeomorphic Deformed " + cfg["m_ScalarMeasurement"],
        "Diffeomorphic Deformation field to Affine space",
        "Diffeomorphic Deformation field to Affine space",
        "Diffeomorphic DTI",
        "Diffeomorphic Deformation field to Original space",
        "DTI-Reg Final DTI"
        ]
    header+=tmp
    with open(outpath,"w") as f:
        csvwriter=csv.writer(f,delimiter=',')
        csvwriter.writerow(header)
        for idx,case in enumerate(zip(cfg["m_CasesIDs"],cfg["m_CasesPath"])):
            caseID,casePath = case
            row=[
                idx+1,
                casePath]
            if m_NeedToBeCropped==1: row+=[m_OutputPath+"/1_Affine_Registration/" + caseID+"_croppedDTI.nrrd"]
            row+=[
                m_OutputPath+"/1_Affine_Registration/" + caseID + "_" + m_ScalarMeasurement + ".nrrd",
                m_OutputPath+"/1_Affine_Registration/Loop" + str(m_nbLoops) + "/" + caseID + "_Loop" + str(m_nbLoops)+"_LinearTrans.txt",
                m_OutputPath+"/1_Affine_Registration/Loop" + str(m_nbLoops) + "/" + caseID + "_Loop" + str(m_nbLoops)+"_LinearTrans_DTI.nrrd",
                m_OutputPath+"/1_Affine_Registration/Loop" + str(m_nbLoops) + "/" + caseID + "_Loop" + str(m_nbLoops)+"_Final" + m_ScalarMeasurement +".nrrd",
                m_OutputPath+"/2_NonLinear_Registration/" + caseID + "_NonLinearTrans_" + m_ScalarMeasurement + ".mhd",
                m_OutputPath+"/2_NonLinear_Registration/" + caseID + "_HField.mhd" ,
                m_OutputPath+"/2_NonLinear_Registration/" + caseID + "_InverseHField.mhd" ,
                m_OutputPath+"/3_Diffeomorphic_Atlas/" + caseID + "_DiffeomorphicDTI.nrrd",
                m_OutputPath+"/4_Final_Resampling/FinalDeformationFields/" + caseID + "_GlobalDisplacementField.nrrd",
                m_OutputPath+"/4_Final_Resampling/FinalTensors/" + caseID + "_FinalDeformedDTI.nrrd"
            ]
            csvwriter.writerow(row)







#!python

import sys
import os 
import traceback
import shutil
from pathlib import Path
import argparse
from argparse import RawTextHelpFormatter
import yaml
sys.path.append(Path(__file__).resolve().parent.parent.__str__()) ## this line is for development
import dtiplayground.dmri.common as common
from dtiplayground.config import INFO as info
from dtiplayground.dmri.preprocessing.app import DMRIPrepApp

logger=common.logger.write 
color= common.Color

def _parse_global_variables(global_vars: list):
    gv = {}
    if global_vars is not None:
        n_vars = int(len(global_vars)/2)
        for i in range(n_vars):
            gv[global_vars[i*2]]=global_vars[i*2+1]
    return gv

##################### commands
def command_init(args):
    ## reparametrization
    options = {
        'config_dir' : args.config_dir,
        'log' : args.log,
        'execution_id' : args.execution_id,
        'no_verbosity' : args.no_verbosity,
        'no_log_timestamp' : args.no_log_timestamp,
        'version' : args.version,
        'tools_dir' : args.tools_dir,
    }
    app = DMRIPrepApp(options['config_dir'])
    return app.initialize(options)
    
def command_remove_module(args):
    options = {
        'config_dir' : args.config_dir,
        'name' : args.name
    }
    app = DMRIPrepApp(options['config_dir'])
    app.remove_module(options)

    return True

def command_add_module(args):
    options = {
        'config_dir': args.config_dir,
        'name' : args.name,
        'base_module' : args.base_module,
        'edit' : args.edit
    }
    app = DMRIPrepApp(options['config_dir'])
    app.add_module(options)


def command_make_protocols(args):
    ## reparametrization
    options={
        "config_dir": args.config_dir,
        "input_images" : args.input_images,
        "module_list": args.module_list,
        "output" : args.output,
        "b0_threshold" : args.b0_threshold,
        "output_format" : args.output_format,
        "no_output_image" : args.no_output_image,
        "global_variables" : _parse_global_variables(args.global_variables)
    }
    app = DMRIPrepApp(options['config_dir'])
    app.makeProtocols(options)


def command_run(args):
    ## reparametrization
    options={
        "config_dir" : args.config_dir,
        "input_image_paths" : args.input_image_list,
        "protocol_path" : args.protocols,
        "output_dir" : args.output_dir,
        "default_protocols":args.default_protocols,
        "num_threads":args.num_threads,
        "execution_id":args.execution_id,
        "baseline_threshold" : args.b0_threshold,
        "output_format" : args.output_format,
        "output_file_base" : args.output_file_base,
        "no_output_image" : args.no_output_image,
        "global_variables" : _parse_global_variables(args.global_variables)
    }
    app = DMRIPrepApp(options['config_dir'])
    app.run(options)
### Arguments 

def get_args():
    version = info['dmriprep']['version']
    logger("VERSION : {}".format(str(version)))
    config_dir=Path.home().joinpath('.niral-dti').resolve()
    # ## read template
    module_help_str=None
    if config_dir.exists() and config_dir.joinpath('config.yml').exists() and config_dir.joinpath('environment.yml').exists():
        config,environment = load_configurations(str(config_dir))
        template_path=config_dir.joinpath(config['protocol_template_path'])
        template=yaml.safe_load(open(template_path,'r'))
        available_modules=template['options']['execution']['pipeline']['candidates']
        available_modules_list=["{}".format(x['value'])  for x in available_modules if x['description']!="Not implemented"]
        module_help_str="Avaliable Modules := \n" + " , ".join(available_modules_list)
    uid, ts = common.get_uuid(), common.get_timestamp()

    ### Argument parsers

    parser=argparse.ArgumentParser(prog="dmriprep",
                                   formatter_class=RawTextHelpFormatter,
                                   description="dmriprep is a tool that performs quality control over diffusion weighted images. Quality control is very essential preprocess in DTI research, in which the bad gradients with artifacts are to be excluded or corrected by using various computational methods. The software and library provides a module based package with which users can make his own QC pipeline as well as new pipeline modules.",
                                   epilog="Written by SK Park (sangkyoon_park@med.unc.edu) , Johanna Dubos (johannadubos32@gmail.com) , Neuro Image Research and Analysis Laboratories, University of North Carolina @ Chapel Hill , United States, 2021")
    subparsers=parser.add_subparsers(help="Commands")
    
    ## init command
    parser_init=subparsers.add_parser('init',help='Initialize configurations')
    parser_init.set_defaults(func=command_init)

    ## add new module
    parser_new_module=subparsers.add_parser('add-module', help='Add new module to user module directory')
    parser_new_module.add_argument('name', help="Module name")
    parser_new_module.add_argument('-b','--base-module', help="Fork from an existing module with new name", default=None, required=False)
    parser_new_module.add_argument('-e','--edit', help="Run vi editor after generating module", default=False, action="store_true")
    parser_new_module.set_defaults(func=command_add_module)

    ## remove user module 
    parser_remove_module=subparsers.add_parser('remove-module', help='Remove a user module from user module directory')
    parser_remove_module.add_argument('name', help="Module name")
    parser_remove_module.set_defaults(func=command_remove_module)

    ## generate-default-protocols
    parser_make_protocols=subparsers.add_parser('make-protocols',help='Generate default protocols',epilog=module_help_str)
    parser_make_protocols.add_argument('-i','--input-images',help='Input image paths',type=str,nargs='+',required=True)
    parser_make_protocols.add_argument('-g','--global-variables',help='Global Variables',type=str,nargs='*',required=False)
    parser_make_protocols.add_argument('-o','--output',help='Output protocol file(*.yml)',type=str)
    parser_make_protocols.add_argument('-d','--module-list',metavar="MODULE",
                                        help='Default protocols with specified list of modules, only works with default protocols. Example : -d DIFFUSION_Check SLICE_Check',
                                        default=None,nargs='*')
    parser_make_protocols.add_argument('-b','--b0-threshold',metavar='BASELINE_THRESHOLD',help='b0 threshold value, default=10',default=10,type=float)
    parser_make_protocols.add_argument('-f','--output-format',metavar='OUTPUT FORMAT',default=None,help='OUTPUT format, if not specified, same format will be used for output (NRRD | NIFTI)',type=str)
    parser_make_protocols.add_argument('--no-output-image',help="No output Qced file will be generated",default=False,action='store_true')
    parser_make_protocols.set_defaults(func=command_make_protocols)
        

    ## run command
    parser_run=subparsers.add_parser('run',help='Run pipeline',epilog=module_help_str)
    parser_run.add_argument('-i','--input-image-list',help='Input image paths',type=str,nargs='+',required=True)
    parser_run.add_argument('-g','--global-variables',help='Global Variables',type=str,nargs='*',required=False)
    parser_run.add_argument('-o','--output-dir',help="Output directory",type=str,required=True)
    parser_run.add_argument('--output-file-base', help="Output filename base", type=str, required=False)
    parser_run.add_argument('-t','--num-threads',help="Number of threads to use",default=1,type=int,required=False)
    parser_run.add_argument('--no-output-image',help="No output Qced file will be generated",default=False,action='store_true')
    parser_run.add_argument('-b','--b0-threshold',metavar='BASELINE_THRESHOLD',help='b0 threshold value, default=10',default=10,type=float)
    parser_run.add_argument('-f','--output-format',metavar='OUTPUT FORMAT',default=None,help='OUTPUT format, if not specified, same format will be used for output  (NRRD | NIFTI)',type=str)
    run_exclusive_group=parser_run.add_mutually_exclusive_group()
    run_exclusive_group.add_argument('-p','--protocols',metavar="PROTOCOLS_FILE" ,help='Protocol file path', type=str)
    run_exclusive_group.add_argument('-d','--default-protocols',metavar="MODULE",help='Use default protocols (optional : sequence of modules, Example : -d DIFFUSION_Check SLICE_Check)',default=None,nargs='*')
    parser_run.set_defaults(func=command_run)

    ## log related
    parser.add_argument('--config-dir',help='Configuration directory',default=str(config_dir))
    parser.add_argument('--log',help='log file',default=str(config_dir.joinpath('log.txt')))
    parser.add_argument('--execution-id',help='execution id',default=uid,type=str)
    parser.add_argument('--no-log-timestamp',help='Remove timestamp in the log', default=False, action="store_true")
    parser.add_argument('--no-verbosity',help='Do not show any logs in the terminal', default=False, action="store_true")
    parser.add_argument('-v','--version', help="Show version", default=False,action="store_true")
    parser.add_argument('--tools-dir', help="Initialize with specific tool directory", default=None)
    ## if no parameter is furnished, exit with printing help
    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args=parser.parse_args()
    if args.version:
        sys.exit(1)

    return args 


## threading environment
args=get_args()
if hasattr(args,'num_threads'):
    os.environ['OMP_NUM_THREADS']=str(args.num_threads) ## this should go before loading any dipy function. 
    os.environ['ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS'] = str(args.num_threads) ## for ANTS threading

import dtiplayground.dmri.preprocessing
import dtiplayground.dmri.preprocessing.modules
import dtiplayground.dmri.preprocessing.protocols

if __name__=='__main__':
    try:
        common.logger.setTimestamp(True)
        result=args.func(args)
        exit(0)
    except Exception as e:
        common.logger.setVerbosity(True)
        msg=traceback.format_exc()
        logger(msg,color.ERROR)
        exit(-1)
    finally:
        pass





from PyQt5.QtWidgets import * 
from PyQt5 import QtCore, QtGui
from PyQt5.QtGui import * 
from PyQt5.QtCore import * 

from dtiplayground.ui.modules_details_communicate import ModulesDetailsCommunicate

class Singletract(QWidget):
  communicate = ModulesDetailsCommunicate()

  def __init__(self, protocol_template, singletract_yml):
    QWidget.__init__(self)
    self.singletract_yml = singletract_yml
    self.method_it = 0
    self.scalar_it = 0
    self.stack = QWidget()
    self.SingletractStack(protocol_template)

  def SingletractStack(self, protocol_template):

    ## Module
    self.tab_name = QLabel()

    ## Description
    for template_ite in range(len(protocol_template["options"]["execution"]["pipeline"]["candidates"])):
      if protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["value"] == "SINGLETRACT_Process":
        description_label = QLabel(protocol_template["options"]["execution"]["pipeline"]["candidates"][template_ite]["description"])
    description_label.setWordWrap(True)

    ## Options:    
    options_groupbox = QGroupBox("Options:")
    options_layout = QGridLayout()
    options_groupbox.setLayout(options_layout)    
    # overwrite
    self.overwrite = QCheckBox("Overwrite")    
    self.overwrite.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.overwrite, 0, 0)
    # skip
    self.skip = QCheckBox("Skip")
    self.skip.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.skip, 0, 1)
    # write image
    self.writeimage = QCheckBox("Write Image")
    self.writeimage.stateChanged.connect(self.GetParams)
    options_layout.addWidget(self.writeimage, 0, 2)
    
    ## Protocol:
    protocol_groupbox = QGroupBox("Protocol:")
    protocol_layout = QFormLayout()
    protocol_groupbox.setLayout(protocol_layout)
    # method
    self.method = QComboBox()
    for ite in self.singletract_yml["protocol"]["method"]["candidates"]:
      self.method.addItem(ite["caption"])
      self.method.setItemData(self.method.count()-1, ite["description"], QtCore.Qt.ToolTipRole)
    self.method.currentTextChanged.connect(self.GetMethodIt)
    self.method.setStatusTip(self.singletract_yml["protocol"]["method"]["description"])
    protocol_layout.addRow(self.singletract_yml["protocol"]["method"]["caption"], self.method)
    self.method.currentTextChanged.connect(self.GetParams)
    # scalar
    self.scalar = QComboBox()
    for ite in self.singletract_yml["protocol"]["scalar"]["candidates"]:
      self.scalar.addItem(ite["caption"])
      self.scalar.setItemData(self.scalar.count()-1, ite["description"], QtCore.Qt.ToolTipRole)
    self.scalar.currentTextChanged.connect(self.GetScalarIt)
    #self.scalar.setStatusTip(self.singletract_yml["protocol"]["scalar"]["description"])
    protocol_layout.addRow(self.singletract_yml["protocol"]["scalar"]["caption"], self.scalar)
    self.scalar.currentTextChanged.connect(self.GetParams)
    # NIRALUtilitiesPath
    self.NIRALUtilitiesPath = QLineEdit()
    self.NIRALUtilitiesPath.setAlignment(Qt.AlignRight)
    self.NIRALUtilitiesPath.setStatusTip(self.singletract_yml["protocol"]["NIRALUtilitiesPath"]["description"])
    protocol_layout.addRow(self.singletract_yml["protocol"]["NIRALUtilitiesPath"]["caption"], self.NIRALUtilitiesPath)
    self.NIRALUtilitiesPath.textChanged.connect(self.GetParams)
    # referenceTractFile
    self.referenceTractFile = QLineEdit()
    self.referenceTractFile.setAlignment(Qt.AlignRight)
    self.referenceTractFile.setStatusTip(self.singletract_yml["protocol"]["referenceTractFile"]["description"])
    protocol_layout.addRow(self.singletract_yml["protocol"]["referenceTractFile"]["caption"], self.referenceTractFile)
    self.referenceTractFile.textChanged.connect(self.GetParams)
    # displacementFieldFile
    self.displacementFieldFile = QLineEdit()
    self.displacementFieldFile.setAlignment(Qt.AlignRight)
    self.displacementFieldFile.setStatusTip(self.singletract_yml["protocol"]["displacementFieldFile"]["description"])
    protocol_layout.addRow(self.singletract_yml["protocol"]["displacementFieldFile"]["caption"], self.displacementFieldFile)
    self.displacementFieldFile.textChanged.connect(self.GetParams)
    # dilationRadius
    self.dilationRadius = QSpinBox()
    self.dilationRadius.setAlignment(Qt.AlignRight)
    self.dilationRadius.setStatusTip(self.singletract_yml["protocol"]["dilationRadius"]["description"])
    protocol_layout.addRow(self.singletract_yml["protocol"]["dilationRadius"]["caption"], self.dilationRadius) 
    self.dilationRadius.valueChanged.connect(self.GetParams)
    
    ## Layout
    layout_v = QVBoxLayout()
    layout_v.addWidget(self.tab_name)
    layout_v.addWidget(description_label)
    layout_v.addWidget(options_groupbox)
    layout_v.addWidget(protocol_groupbox)
    layout_v.addStretch(1)
    self.stack.setLayout(layout_v)

  
  def GetMethodIt(self, text): # dti register parameter
    for it in range(len(self.singletract_yml["protocol"]["method"]["candidates"])):
      if self.singletract_yml["protocol"]["method"]["candidates"][it]["caption"] == text:
        self.method_it = it

  def GetScalarIt(self, text): # dti register parameter
    for it in range(len(self.singletract_yml["protocol"]["scalar"]["candidates"])):
      if self.singletract_yml["protocol"]["scalar"]["candidates"][it]["caption"] == text:
        self.scalar_it = it

  def GetSimilarityMetricIt(self, text): # dti register parameter
    for it in range(len(self.singletract_yml["protocol"]["similarityMetric"]["candidates"])):
      if self.singletract_yml["protocol"]["similarityMetric"]["candidates"][it]["caption"] == text:
        self.similarityMetric_it = it


  def GetParams(self):
    params = [
      'Register DTI (ANTs)',
      'SINGLETRACT_Process', 
      {'options': {
        'overwrite': self.overwrite.isChecked(), 
        'skip': self.skip.isChecked(), 
        'write_image': self.writeimage.isChecked()
        }, 
      'protocol': {
        'method': self.singletract_yml["protocol"]["method"]["candidates"][self.method_it]["value"], 
        'scalar': self.singletract_yml['protocol']['scalar']['candidates'][self.scalar_it]['value'],
        'NIRALUtilitiesPath': self.NIRALUtilitiesPath.text(),
        'referenceTractFile': self.referenceTractFile.text(), 
        'displacementFieldFile': self.displacementFieldFile.text(),
        'dilationRadius': self.dilationRadius.value()
        }
      }
    ]
    self.communicate.SendParams(params)


#!python

import shutil
import os 
import importlib
from pathlib import Path
import traceback
import time
import yaml
import uuid
import sys
import re

from dtiplayground.config import INFO as info
import dtiplayground
import dtiplayground.dmri.common as common
import dtiplayground.dmri.common.module as module

logger=common.logger.write 
color= common.Color

class AppBase:
    def __init__(self,config_root, app_name,*args,**kwargs):
        self.config_dir = Path(config_root)
        self.app = {}
        self.app['application'] = app_name
        self.getAppInfo()
        if 'application_dir' in self.app:
            self.log_file = str(Path(self.app['application_dir']).joinpath('log.txt'))
            self.app.setdefault('no_log_timestamp',False)
            self.app.setdefault('log', self.log_file)
            self.app.setdefault('no_verbosity', False)
            self.app.setdefault('execution_id',  common.get_uuid())
            self.kwargs=kwargs
        else:
            self.kwargs=kwargs
        self.globalvars = {}
        logger(yaml.dump(self.app))

    def getAppInfo(self):
        version=info[self.app['application']]['version']
        application_dir=Path(self.config_dir).joinpath('{}-{}'.format(self.app['application'],str(version))).resolve()
        res = {
            'version' : version,
            'config_dir' : str(self.config_dir),
            'application_dir' : str(application_dir)
        }
        self.app.update(res)
        self.getAppInfoImpl()

        return self.app 

    def getAppInfoImpl(self):
        return self.app

    def initialize(self, options):
        ## reparametrization
        _args = {
            'log' : self.app['log'],
            'execution_id' : self.app['execution_id'],
            'no_verbosity' : self.app['no_verbosity'],
            'no_log_timestamp' : self.app['no_log_timestamp'],
        }
        _args.setdefault('tools_dir',None)

        home_dir = Path(self.app['application_dir'])
        version = self.app['version']

        ## Function begins
        home_dir.mkdir(parents=True,exist_ok=True)
        tools_dir = None
        if _args['tools_dir'] is not None:
            tools_dir = Path(_args['tools_dir'])

        globalvars_fn=Path(self.app['config_dir']).joinpath('global_variables.yml')

        if globalvars_fn.exists():
            self.globalvars = yaml.safe_load(open(globalvars_fn,'r'))

        if tools_dir is not None:
            if tools_dir.exists():
                dptdir = tools_dir.joinpath('dtiplayground-tools')
                fsldir = tools_dir.joinpath('FSL')
                self.globalvars.update({
                    "dtiplayground-tools" :{
                        "path" : dptdir.resolve().__str__(),
                        "info" : yaml.safe_load(open(dptdir.joinpath('info.yml'),'r'))
                    },
                    "fsl" : {
                        "path" : fsldir.resolve().__str__(),
                        "info" : {
                            'name' : 'fsl',
                            "verson" : open(fsldir.joinpath('etc/fslversion'),'r').read().strip()
                        }
                    }
                })
                yaml.dump(self.globalvars, open(globalvars_fn,'w'))
            else:
                logger("Couldn't find tool directory {}".format(str(tools_dir)),color.ERROR)
                exit(1)

        self.initialize_logger(_args)

        self.initializeImpl(options)

        initialization_fn = Path(self.app['application_dir']).joinpath("init.yml")
        yaml.dump(self.app, open(initialization_fn,'w'))
        logger("Initialized. Local configuration will be stored in {}".format(str(home_dir)),color.OK)

    def initializeImpl(self, options):
        pass

    def run(self,options): ### options must have output_dir 
        return self.runImpl(options)

    def runImpl(self, options):
        return True

    ################################################
    ## decorators (only used in the memeber function)
    #################################################

    def after_initialized(self,func): #decorator for other functions other than command_init
        def wrapper(args):
            home_dir=Path(self.app['application_dir'])
            if home_dir.exists() and home_dir.joinpath('init.yml').exists():
                logger("Configuration directory is found : {}".format(self.app['application_dir']),color.OK)
                config_file=home_dir.joinpath('config.yml')
                environment_file=home_dir.joinpath('environment.yml')
                self.initialize_logger(self.app)
                return func(args)
            else: 
                # raise Exception("Not initialized, please run init command at the first use")
                logger("Configuration directory is not found or some files are missing, commencing initialization ...",color.WARNING)
                self.initialize(args)
                config_file=home_dir.joinpath('config.yml')
                environment_file=home_dir.joinpath('environment.yml')
                self.initialize_logger(self.app)
                return func(args)
        return wrapper

    #################################################
    ######## utilities
    #################################################

    def initialize_logger(self,_args):
        ## default log setting
        common.logger.setLogfile(_args['log'])
        common.logger.setTimestamp(not _args['no_log_timestamp'])
        common.logger.setVerbosity(not _args['no_verbosity'])


    def _update_software_paths(self, globalvars):
        home_dir=Path(self.app['application_dir'])
        software_info_path=Path(dtiplayground.__file__).resolve().parent.joinpath("dmri/common/data/software_paths.yml")
        software_paths=yaml.safe_load(open(software_info_path,'r'))
        software_paths=self._resolve_softwarepaths(software_paths, globalvars)
        software_filename=home_dir.joinpath('software_paths.yml')
        yaml.dump(software_paths, open(software_filename,'w'))
    
    def _resolve_softwarepaths(self,spathobj, globalvars):
        if 'dtiplayground-tools' in globalvars:
            p = Path(globalvars['dtiplayground-tools']['path'])
            if p.exists():
                os.environ['DTIPLAYGROUNDTOOLS'] = p.resolve().__str__()

        if 'fsl' in globalvars:
            p = Path(globalvars['fsl']['path'])
            if p.exists():
                os.environ['FSL'] = p.resolve().__str__()

        if  'DTIPLAYGROUNDTOOLS' not in os.environ:
            os.environ['DTIPLAYGROUNDTOOLS']=os.path.expandvars("$HOME/.niral/dtiplayground-tools")
            tooldir=os.environ['DTIPLAYGROUNDTOOLS']
            if Path(tooldir).exists():
                logger("DTI Playground tools directory found at {}".format(tooldir), color.OK)
            else:
                logger("DTI Playground tools directory is set to {}".format(tooldir), color.WARNING)
                logger("DTI Playground tools directory not found, please install toolkits or specify software paths manually", color.WARNING)
        else:
            tooldir=os.environ['DTIPLAYGROUNDTOOLS']
            if Path(os.environ['DTIPLAYGROUNDTOOLS']).exists():
                
                logger("DTI playground tools are found in {}".format(tooldir),color.OK)
            else:
                logger("DTI playground tools are not found in {}".format(tooldir),color.ERROR)
        if 'FSL' not in os.environ:
            logger("FSL directory is not set in environment variable FSL",color.WARNING)
            logger("Make sure FSL directory is set in the software paths",color.WARNING)
        else:
            if Path(os.environ['FSL']).exists():
                logger("FSL directory is found at {}".format(os.environ['FSL']),color.OK)
            else:
                logger("FSL directory is set in environment variable FSL, but doesn't exist",color.WARNING)
        
        sp = spathobj['softwares']
        for k,v in sp.items():
            resolved_path = os.path.expandvars(v['path'])
            sp[k]['path'] = resolved_path
        spathobj['softwares'] = sp
        return spathobj


